/* pa_main.c (Enumerate all partitions of a set: main program)
   
   This is equivalent to enumerating all semilabeled trees.
   (see Peter L. Erd"os, Enumeration of rooted forests, Report No. 90631 - OR,
   Forschungsinstitut fuer diskrete Mathematik, Rheinische Friedrich-Wilhelms-
   Universitaet, Bonn 1990, ETH-BIB P 715070:90631) "

   See also help(bell) in Maple.
   
   A backtrack application in ZRAM
   Ambros Marzetta

   n	solutions
   1   	        1
   2	        2
   3	        5
   4	       15
   5	       52
   6	      203
   7	      877
   8	     4140
   9	    21147
   10	   115975
   11	   678570
   12     4213597

   The branching procedure is small, thus the ZRAM overhead is relatively
   large.
   
 * $Source: /home/schilte/marzetta/work/distr/zram/RCS/pa_main.c,v $
 * $Revision: 1.1 $
 * $Date: 1998/04/15 18:36:04 $
 * $Log: pa_main.c,v $
 * Revision 1.1  1998/04/15  18:36:04  marzetta
 * Initial revision
 *
 */

#include <stdlib.h>		/* exit() */
#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include "z_base.h"
#include "z_bt.h"
#include "z_par.h"		/* z_myproc */
#include "pa_base.h"
#include "z_time.h"
#include "z_lb.h"		/* z_lb_redundantComputation */


static z_bt_ProcSetT pa_procset1;
static int pa_procsetnum1;
static pa_TotalP total;
#ifndef SEQUENTIAL
static z_JobNum totalJobNum;
#endif /* SEQUENTIAL */

void pa_BranchInPlace(pa_GlobalP global, pa_NodeP this,
			 z_bt_DoProc whatToDo, void *ref)
{
  /* may overwrite this->act_elt */
  int elt, a, new[MaxN], count;

  if (this->unassigned == 0) {	/* found a partition */
    if (!global->quiet) {
#if 0
      printf("[%2d]Solution:", z_myproc);
#endif
      pa_Print(global, this);
    }
    /* count the solutions */
    if (!z_lb_redundantComputation) total->sol_count ++;
    return;
  }
  
  assert(this->b >= 1);
  /*pa_Print(global, this);*/
  if (this->b == 1) {		/* add all remaining elements to one class */
    assert(this->act_elt == 0);
    assert(this->unassigned <= MaxN);
    count = 0;
    for (elt = this->act_elt; elt < global->n; elt++)
      if (pa_class(this, elt) == 0) {
	new[count] = elt;
	count ++;
	pa_class(this, elt) = this->act_num;
      }
    assert(this->unassigned == count);
    this->unassigned = 0;
    /*pa_BranchInPlace(global, this, whatToDo, ref);*/
    whatToDo((z_NodeP) this, ref);
    this->unassigned = count;
    while (count > 0) {
      -- count;
      pa_class(this, new[count]) = 0;
    }
    return;
  }
  
  elt = this->act_elt;
  if (elt == 0) {		/* add the first element to a class */
    while (pa_class(this, elt) != 0) elt++;
    pa_class(this, elt) = this->act_num;
    this->act_elt = elt + 1;
    this->unassigned --;
    /*pa_BranchInPlace(global, this, whatToDo, ref);*/
    whatToDo((z_NodeP) this, ref);
    this->unassigned ++;
    pa_class(this, elt) = 0;
    return;
  }

  assert(this->a <= this->unassigned + 1);
  while (elt < global->n && pa_class(this, elt) != 0) elt++;
  if (elt < global->n && this->a < this->unassigned + 1) {
    this->act_elt = elt + 1;
    whatToDo((z_NodeP) this, ref);	/* without elt */
    pa_class(this, elt) = this->act_num;
    this->act_elt = elt + 1;
    this->unassigned --;
    whatToDo((z_NodeP) this, ref);	/* with elt */
    this->unassigned ++;
    pa_class(this, elt) = 0;
  } else {
    a = this->a;
    if (this->a) this->a --;
    this->b --;
    this->act_num ++;
    this->act_elt = 0;
    whatToDo((z_NodeP) this, ref);
    this->a = a;
    this->b ++;
    this->act_num --;
  }
  return;
} /* pa_BranchInPlace */

#ifndef SEQUENTIAL
void CollectTotalJob(void)
{
  /* This procedure, which is called on every processor at the end of the
     computation, sends the local count of solutions back to the front end.
     */
  z_AddNodeBuffer((z_NodeP)total);
} /* CollectTotalJob */

void AddTotal(int pe)
{
  /* This procedure is executed on the front end. */
  pa_TotalP t;
  
  if (pe != z_myproc) {
    t = (pa_TotalP) z_GetNodeBuffer();
    total->sol_count += t->sol_count;
  }
} /* AddTotal */
#endif /* SEQUENTIAL */

void initproc(void)
{
  /* This function is called once on every processing element. */

  z_bt1_Init();
  z_bt2_Init();
  pa_base_init();
  pa_procset1.bt_BranchInPlace = (z_bt_BranchProc)pa_BranchInPlace;
  pa_procsetnum1 = z_InstallProcSet((z_ProcSetP) &pa_procset1);
#ifndef SEQUENTIAL
  totalJobNum = z_InstallJob(CollectTotalJob);
#endif /* SEQUENTIAL */
  total = (pa_TotalP) z_NewNode(pa_TotalTag, sizeof(pa_Total));
  total->sol_count = 0;
} /* initproc */

void PrintTime(double time, int n, const char *comment)
{
  time = z_Gtime() - time;
  printf("Time:%7.2f (%2d processor%s) n = %d %s\n",
	 time, z_numprocs, z_numprocs > 1 ? "s" : "", n,
	 comment);
} /* PrintTime */

int Usage(const char *name)
{
  fprintf(stderr, "Usage: %s [-1hnt] n [a b]\n", name);
  fprintf(stderr, "       Enumerate all partitions of the set {1..n} into\n");
  fprintf(stderr, "       at least a and at most b disjoint non-empty subsets\n");
  fprintf(stderr, "Options: -1  sequential\n");
  fprintf(stderr, "         -h  help (produce this message)\n");
  fprintf(stderr, "         -n  no output (implies time measurement)\n");
  fprintf(stderr, "         -t  time measurement\n");
#if 0
  fprintf(stderr, "Output: 1 1 2 1 2 means there are two subsets,\n");
  fprintf(stderr, "        subset 1 is {1,2,4} and subset 2 is {3,5}\n");
#endif
  return 4;
} /* Usage */

int mmain(int argc, char **argv, char **envp)
{
  pa_GlobalP global;
  pa_NodeP root;
  int i;
  double time;
  boolean showTime=0;
  boolean sequential=0;
  boolean noOutput=0;
  
  char c;

  while ((c=getopt(argc, argv, "1chnprt")) != -1)
    switch(c) {
    case 'n':
      noOutput = 1;
      showTime = 1;
      break;
    case 't':
      showTime = 1;
      break;
    case '1':
      sequential = 1;
      break;
    default:
      return Usage(argv[0]);
    }
  
  if (argc < optind+1) return Usage(argv[0]);
  global = (pa_GlobalP) z_NewNode(pa_GlobalTag, sizeof(pa_Global));
  if (!global) exit(1);
  global->n = atoi(argv[optind]);
  if (global->n < 1 || global->n > MaxN) return Usage(argv[0]);
  global->quiet = noOutput;
  
  root = pa_NewNode(global->n);
  if (argc == optind + 3) {
    root->a = atoi(argv[optind+1]);
    root->b = atoi(argv[optind+2]);
    if (root->b < 1 || root->a > global->n || root->b < root->a) {
      printf("No solutions for this a and b\n");
      return 1;
    }
  } else {
    root->a = 1;
    root->b = global->n;
    if (argc != optind + 1) return Usage(argv[0]);
  }
  root->act_num = 1;
  root->act_elt = 0;
  root->unassigned = global->n;
  for (i=0; i<global->n; i++) pa_class(root, i) = 0;
  
  time = z_Gtime();
  z_bt_EnumerateDFExp(pa_procsetnum1,
		      (z_NodeP)global, (z_NodeP)root,
		      sequential);
  if (showTime) PrintTime(time, global->n, "");
#ifndef SEQUENTIAL
  /* Add the solution counters of all processors. */
  z_StartParallelJob(totalJobNum, AddTotal);
  z_WaitWork();
#endif /* SEQUENTIAL */
  fprintf(stderr, "Number of solutions: %ld\n", total->sol_count);
  return 0;
} /* mmain */
