/* z_base.c (ZRAM: base definitions)
   Ambros Marzetta

 * $Source: /home/schilte/marzetta/work/distr/zram/RCS/z_base.c,v $
 * $Revision: 1.8 $
 * $Date: 1996/06/03 19:00:37 $
 * $Log: z_base.c,v $
 * Revision 1.8  1996/06/03  19:00:37  marzetta
 * New definition of z_NewNode.
 *
 * Revision 1.7  1996/05/29  11:32:56  marzetta
 * eliminated compiler warnings
 *
 * Revision 1.6  1996/04/09  13:20:21  marzetta
 * new Install procedures
 *
 * Revision 1.5  1995/09/14  12:25:56  marzetta
 * deep copy
 *
 * Revision 1.4  1995/08/22  12:39:12  marzetta
 * New message passing, MPI
 *
 */

#include <stdio.h>
#include <stdlib.h> /* exit() */
#include <string.h> /* memcpy() */
#include "z_base.h"

/* Interrupts ----------------------------------------------------------- */

volatile boolean z_terminate;
/* Will be set when the process receives a terminate signal. */

volatile boolean z_checkpoint;
/* Will be set when some timer expires. */

/* Installing classes --------------------------------------------------- */

#define MAXCLASS Z_NUM_TAGS
#define NUMCLASSSTART 20	/* should be 1 */

static int numClass = NUMCLASSSTART;
static const char *className[MAXCLASS];
static int classChildren[MAXCLASS];

int z__InstallClass(const char *name, int children,
		    int reserved1, int reserved2)
/* Install a node type. z__InstallClass returns a tag to be used in
   subsequent calls to z_NewNode and in accesses to z_numNodes. This is
   a collective operation. */
{
  assert(reserved1 == 0);
  assert(reserved2 == 0);
  assert(numClass < MAXCLASS);
  className[numClass] = name;
  classChildren[numClass] = children;
  /*printf("Installing %s with tag %d and %d children\n", name, numClass, children);*/
  return numClass ++;
} /* z_InstallClass */

/* Memory management ---------------------------------------------------- */

#define FREELISTSIZE 10

typedef struct z_FamilyNode {
  Z_NODE_HEADER;
  z_NodeP ch[10];
} *z_FamilyNodeP;

long z_numNodes[Z_NUM_TAGS];
static struct z_Node freeList;
int freeSize;

static z_FamilyNodeP z_Allocate(MINT len)
{
  z_NodeP np, f;

  z_memavail -= len;
  np = NULL;
  for (f = &freeList; f->child; f = f->child) {
    if (f->child->len == len) {
      np = f->child;
      f->child = np->child;
      freeSize --;
      return (z_FamilyNodeP) np;
    }
  }
  return (z_FamilyNodeP) malloc(len);
} /* z_Allocate */

z_NodeP z_NewNode(MINT tag, MINT len)
/* Allocate a z_Node, initialize tag, len, children */
{
  z_FamilyNodeP np;

  assert(0 < tag && tag < Z_NUM_TAGS);
  /* forbid tag == 0 to detect uninitialized tags */
  assert(className[tag]);
  /*printf("Allocating %ld bytes (tag = %ld)\n", (long)len, (long)tag);*/
  np = z_Allocate(len);
  if (np) {
    np->len = len;
    np->tag = tag;
    np->children = classChildren[tag];
    np->parent = NULL;
    np->child = NULL;
    np->sibling = NULL;
    z_numNodes[tag] ++;
  }
  if (! np) {			/* where is the memory lost? */
    int i;
    z_NodeP f;
    
    printf("Allocated nodes:\n");
    for (i=NUMCLASSSTART; i<numClass; i++)
      printf("  %-20s %4ld\n", className[i], z_numNodes[i]);
    printf("Free list:\n");
    for (f = freeList.child; f; f = f->child)
      printf("  %-20s      %d\n", className[f->tag], f->len);
    fflush(stdout);
  }
  return (z_NodeP) np;
} /* z_NewNode */

z_NodeP z_NewCopyNode(z_NodeP source)
{
  /* Allocate a z_Node identical to source (deep copy) */
  z_FamilyNodeP src, dest;
  int i;
  
  assert(source);
  src = (z_FamilyNodeP) source;
  dest = z_Allocate(source->len);
  if (! dest) {			/* where is the memory lost? */
    int i;
    z_NodeP f;
    
    printf("Allocated nodes:\n");
    for (i=NUMCLASSSTART; i<numClass; i++)
      printf("  %-20s %4ld\n", className[i], z_numNodes[i]);
    printf("Free list:\n");
    for (f = freeList.child; f; f = f->child)
      printf("  %-20s      %d\n", className[f->tag], f->len);
  }
  if (dest) {
    z_numNodes[source->tag] ++;
    memcpy((char *) dest, (char *) source, source->len);
    for (i=src->children-1; i>=0; --i)
      if (src->ch[i]) dest->ch[i] = z_NewCopyNode(src->ch[i]);
  } else {
    printf("z_NewCopyNode: no memory (%ld nodes allocated)\n",
	   z_numNodes[source->tag]);
    fflush(stdout);
  }
  assert(dest);
  return (z_NodeP) dest;
} /* z_NewCopyNode */

void z_FreeNode(z_NodeP np)
{
  z_FamilyNodeP f;
  int i;
  
  f = (z_FamilyNodeP) np;
  for (i=f->children-1; i>=0; --i)
    if (f->ch[i]) z_FreeNode(f->ch[i]);
  z_numNodes[np->tag] --;
  z_memavail += np->len;
  if (freeSize < FREELISTSIZE) {
    f->child = freeList.child;
    freeList.child = (z_NodeP) f;
    freeSize ++;
  } else
    free(np);			/* z_FreeNode */
} /* z_FreeNode */

void z_ReSizePointer(void **ptr, long *size, long newsize)
{
  /* Allocate if *size < newsize. Don't copy any data. */
  if (newsize > *size) {
    if (*ptr)
      (void) free(*ptr);	/* z_ReSizePointer */
    *size = newsize;
    *ptr = (void *) malloc(newsize); /* z_ReSizePointer */
    assert(*ptr);
  }
} /* z_ReSizePointer */

